from trilogy.core.models import (
    Concept,
    Environment,
)
from langchain.tools import Tool, StructuredTool
import json
from trilogy_nlp.tools import get_today_date
from trilogy_nlp.exceptions import ValidationPassedException

from trilogy_nlp.llm_interface.validation import (
    validate_response,
    ValidateResponseInterface,
    VALID_STATUS,
)


def get_model_description(query: str, environment: Environment):
    """
    Get the description of the dataset.
    """
    datasources = [{"name": x.identifier} for x in environment.datasources.values()]
    return json.dumps(
        {
            "description": f"Contains the following datasources: {datasources}. Use the get_fields tool with a name to get more specific information about any given one."
        }
    )


def concept_to_string(concept: Concept) -> str:
    return concept.address


def get_fields(environment: Environment, search: str, *args, **kwargs) -> str:
    return json.dumps(
        {
            "fields": [
                (
                    {
                        "name": concept_to_string(x),
                        "description": x.metadata.description,
                    }
                    if x.metadata.description
                    else {
                        "name": concept_to_string(x),
                    }
                )
                for x in environment.concepts.values()
                if "__preql_internal" not in x.address
                and not x.address.endswith(".count")
            ]
        }
    )


def get_help(x):
    raise SyntaxError(x)


def submit_response(environment, prompt, **kwargs):
    response, ir = validate_response(
        environment=environment,
        prompt=prompt,
        **kwargs,
    )
    if response["status"] == VALID_STATUS:
        raise ValidationPassedException(ir=ir)
    return response


def sql_agent_tools(environment, prompt: str):
    def validate_response_wrapper(**kwargs):
        response, ir = validate_response(
            environment=environment,
            prompt=prompt,
            **kwargs,
        )
        return response

    def submit_wrapper(**kwargs):
        return submit_response(
            environment=environment,
            prompt=prompt,
            **kwargs,
        )

    tools = [
        Tool.from_function(
            func=lambda x: get_model_description(x, environment),
            name="get_validation_help",
            description="""
           Get help with validation errors""",
            handle_tool_error=True,
        ),
        StructuredTool(
            name="validate_response",
            description="""
            Check that a response is formatted properly and accurate before your final answer. Always call this with the complete final response before reporting a Final Answer!
            If the response is not correct, the "valid" argument will be false and it will return an array of errors. If it is correct, it will return "true" for the valid argument.
            You must fix all errors returned by validation before submitting. If you are not sure what the error means, ask for help using the get_validation_help tool.
            """,
            func=validate_response_wrapper,
            args_schema=ValidateResponseInterface,
            handle_tool_error=True,
        ),
        StructuredTool(
            name="submit_answer",
            description="""
            Submit your final answer. It will be validated, so this function may return errors to correct, similar to the validate_response function. If it passes, the answer will be submitted.
            """,
            func=submit_wrapper,
            args_schema=ValidateResponseInterface,
        ),
        Tool.from_function(
            func=lambda x: get_fields(environment, x),
            name="get_fields",
            description="""
            Returns array of json objects containing the names of actual fields that can be referenced, with a description if it exists. Fields always need to be referenced by exact name. Queries operate on fields only, by exact name.
            """,
        ),
        Tool.from_function(
            func=get_today_date,
            name="get_today_date",
            description="""
            Use to get the date of today.
            """,
        ),
    ]
    return tools



from pathlib import Path
from langchain_core.language_models import BaseLanguageModel
from trilogy import Environment
from pydantic import BaseModel
from langchain.agents import create_structured_chat_agent, AgentExecutor
from langchain.tools import Tool, StructuredTool
from trilogy.core.enums import Purpose
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from trilogy_nlp.tools import get_wiki_tool
from trilogy.core.enums import Modifier
from difflib import get_close_matches
from trilogy.core.exceptions import InvalidSyntaxException
from trilogy_nlp.constants import logger

# from trilogy.core.exceptions import Import


class AddImportResponse(BaseModel):
    """result of initial parse request"""

    namespaces: list[str]
    reasoning: str


def get_environment_possible_imports(env: Environment) -> list[str]:
    raw = Path(env.working_path).glob("*.preql")
    return [x.stem for x in raw if "query" not in x.stem]


def get_environment_detailed_values(env: Environment, input: str):
    new = Environment(working_path=env.working_path)
    try:
        new.parse(f"import {input} as {input};")

    except (ImportError, InvalidSyntaxException):
        suggestions = get_close_matches(input, get_environment_possible_imports(env))
        base = "This is an invalid database. Refer back to the list from the get_database_list tool. Names must match exactly."
        if suggestions:
            base += f" Did you mean {suggestions[0]}?"
        return base
    # TODO: does including a description work better?
    return {
        k.split(".", 1)[1]
        for k, v in new.concepts.items()
        if "__preql_internal" not in k and not k.endswith(".count")
    }


def validate_response(namespaces: list[str], reasoning: str, environment: Environment):
    possible = get_environment_possible_imports(environment)
    if not all(x in possible for x in namespaces):
        return {"status": "invalid", "error": "Not all of those databases exist!"}

    return {
        "status": "valid",
    }


def environment_agent_tools(environment):
    def get_import_wrapper(*args, **kwargs):
        return get_environment_possible_imports(environment)

    def validate_response_wrapper(namespaces: list[str], reasoning: str):
        return validate_response(
            namespaces=namespaces, reasoning=reasoning, environment=environment
        )

    tools = [
        Tool.from_function(
            func=get_import_wrapper,
            name="list_databases",
            description="""
           Describe the databases you can look at. Takes empty argument string.""",
            handle_tool_error='Argument is an EMPTY STRING, rendered as "". {} is not valid. Do not call with {} ',
        ),
        Tool.from_function(
            func=lambda x: get_environment_detailed_values(environment, x),
            name="get_database_description",
            description="""
           Describe the database and general groupings of fields available. Call with a database name.""",
            handle_tool_error=True,
        ),
        StructuredTool(
            name="validate_response",
            description="""
            Check that a response is formatted properly and accurate before your final answer. Always call this!
            """,
            func=validate_response_wrapper,
            args_schema=AddImportResponse,
        ),
    ]
    return tools


def llm_loop(
    input_text: str, input_environment: Environment, llm: BaseLanguageModel
) -> AddImportResponse:
    system = """Thought Process: You are a data analyst assistant. Your job is to identify the datasource most relevant to answering a business question. 
    You should select the minimum number of databases that covers information. Some databases will reference others; if that's the case, 
    eg sales.customer; you do not need to explicitly include the customer database as well when looking for customer information about store sales.

    Example: if you are asked for "orders by customer", you ONLY return the orders database.

    If you need additional contextual information to understand the query that would not be in the database, use the wikimedia tool to get it.

    Sometimes you may need multiple sources:

    Example: If you are asked for "address of all employees, and how many orders they've placed" you would return employees and orders, because not all employees may have placed orders. 

    If the question suggests you use a specific datasource, eg "using ocean shipment data", assume that is sufficient.

    The output to the analyst should be a VALID JSON blob with the following keys and values followed by a stopword: <EOD>:
    - namespaces: a list of databases to use
    - reasoning: a string explaining why all are required. Remember, each additional database you include costs money!

    To start, pick a database and call the get_database_description tool on it. This will give you a description of the database and its fields. 
    Continue until you believe you have found all fields required to answer the question.

    You should always call the the validate_response tool on what you think is the final answer before returning the "Final Answer" action.
    You have access to the following tools:

    {tools}

    Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input). 
    You will get valuable information from using tools before producing your final answer.

    Use as many tools as needed before producing the "Final Answer" action.

    Valid "action" values: any of {tool_names} and, for your last result "Final Answer". 
    Only return "Final Answer" when you are done with all work. Nnever set the action to 
    "Final Answer" before you are done. 

    You should always call the the validate_response tool with your final answer before sending it to the CEO.

    Provide only ONE action per $JSON_BLOB, as shown:

    ```
    {{
        "action": $TOOL_NAME,
        "action_input": $INPUT
    }}
    ```

    Follow this format in responses:

    Question: input question to answer
    Thought: consider previous and subsequent steps
    Action:
    ```
    $JSON_BLOB
    ```
    Observation: action result
    ... (repeat Thought/Action/Observation N times)

    Action input is in JSON format, not as a JSON string blob (No escaping!)

    An example series:

    Question: Get all customers who live in US zip code 10245 and how many orders they placed in 2000
    Thought: I should get the description of of the orders dataset
    Action:
    ```
    {{
        "action": "get_database_description",
        "action_input": "orders"
    }}

    Observation: <some description>
    Thought: I should check my answer
    Action:
    ```
    {{
        "action": "validate_response",
        "action_input": "{{
        "namespaces": [
            "orders", "customers"
        ]
    }}"
    }}

    Once you have used any tools (listed below) as needed, you will produce your final result in this format. After producing your
    final answer, you cannot take any more steps.

    A final answer would look like this:
    {{
        "action": "Final Answer",
        "action_input": <VALID_JSON_SPEC_DEFINED_ABOVE>
    }}

    Begin! Reminder to ALWAYS respond with a valid json blob of an action. Always use tools. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation"""

    human = """{input}

    {agent_scratchpad}

    (reminder to respond in a JSON blob no matter what)"""

    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", system),
            MessagesPlaceholder("chat_history", optional=True),
            ("human", human),
        ]
    )

    tools = environment_agent_tools(input_environment) + [get_wiki_tool()]
    chat_agent = create_structured_chat_agent(
        llm=llm,
        tools=tools,
        prompt=prompt,
    )
    agent_executor = AgentExecutor(
        agent=chat_agent, tools=tools, verbose=True, handle_parsing_errors=True  # type: ignore
    )

    result = agent_executor.invoke({"input": input_text})
    output = result["output"]
    if isinstance(output, str):
        return AddImportResponse.model_validate_json(output)
    elif isinstance(output, dict):
        return AddImportResponse.model_validate(output)
    else:
        raise ValueError("Unable to parse LLM response")


def select_required_import(
    input_text: str, environment: Environment, llm: BaseLanguageModel
) -> AddImportResponse:
    exception = None
    ATTEMPTS = 3
    for attempts in range(ATTEMPTS):
        try:
            return llm_loop(input_text, environment, llm)
        except Exception as e:
            logger.error("Error in select_required_import: %s", e)
            exception = e
    if exception:
        raise exception
    raise ValueError(f"Unable to get parseable response after {ATTEMPTS} attempts")


def build_env_and_imports(
    input_text: str,
    working_path: Path,
    llm: BaseLanguageModel,
):
    base = Environment(working_path=working_path)
    response: AddImportResponse = select_required_import(input_text, base, llm=llm)

    for x in response.namespaces:
        base.parse(f"""import {x} as {x};""")

    # automatically merge concepts between the imported domains
    for k, concept in base.concepts.items():
        if not concept.namespace:
            continue
        # only merge on keys automatically
        if not concept.purpose == Purpose.KEY:
            continue
        # don't merge on date
        # TODO: be less hacky in terms of what default merges we should do
        if "date" in concept.namespace:
            continue
        if "__preql_internal" in k:
            continue
        pre_namespace_root = k.split(".", 1)[1]
        if "." not in pre_namespace_root:
            continue
        if pre_namespace_root in base.concepts:
            base.merge_concept(
                concept, base.concepts[pre_namespace_root], modifiers=[Modifier.PARTIAL]
            )

    return base



from enum import Enum


class Provider(Enum):
    OPENAI = "openai"
    OCTO_AI = "octoai"
    LOCAL = "local"
    LLAMAFILE = "llamafile"
    GOOGLE = "google"


class CacheType(Enum):
    MEMORY = "memory"
    SQLLITE = "sqlite"



from typing import Optional
from trilogy.core.models import (
    Environment,
)
from typing import List
from pydantic import BaseModel
from pydantic_core import ValidationError, ErrorDetails
from trilogy.core.enums import ComparisonOperator
from langchain_core.tools import ToolException
from trilogy.core.exceptions import UndefinedConceptException
from collections import defaultdict

from trilogy_nlp.llm_interface.models import (
    InitialParseResponseV2,
    Column,
    NLPConditions,
    FilterResultV2,
    OrderResultV2,
    Literal,
    Calculation,
)
from trilogy_nlp.llm_interface.constants import COMPLICATED_FUNCTIONS
import difflib

# from trilogy.core.constants import
from trilogy.core.enums import (
    FunctionType,
)
from trilogy.core.models import DataType
from enum import Enum
import json

VALID_STATUS = "valid"


def is_valid_function(name: str):
    return name.lower() in [item.value for item in FunctionType]


def invalid_operator_message(operator: str) -> str | None:
    try:
        ComparisonOperator(operator)
    except Exception as e:
        return str(e)
    return None


VALIDATION_CACHE: dict[str, InitialParseResponseV2] = {}


class QueryContext(Enum):
    SELECT = "COLUMN"
    FILTER = "FILTER"
    ORDER = "ORDER"

    def __str__(self) -> str:
        return f"{self.value} definition"


def validate_query(
    query: dict, environment: Environment, prompt: str
) -> tuple[dict, InitialParseResponseV2 | None]:
    try:
        parsed = InitialParseResponseV2.model_validate(query)
    except ValidationError as e:
        return {"status": "invalid", "error": validation_error_to_string(e)}, None
    errors = []
    # assume to start that all our select calculations are valid
    select = {col.name for col in parsed.output_columns if col.calculation}
    filtered_on = set()

    def validate_calculation(calc: Calculation, context: QueryContext) -> bool:
        valid = True
        if not is_valid_function(calc.operator):
            errors.append(
                f"{calc} in {context} does not use a valid function (is using {calc.operator}); check that you are using ONLY a valid option from this list: {sorted([x for x in FunctionType.__members__.keys() if x not in COMPLICATED_FUNCTIONS]) }. If the column requires no transformation, drop the calculation field.",
            )
        if calc.over:
            for x in calc.over:
                local = validate_column(x, context)
                valid = valid and local

        for arg in calc.arguments:
            if isinstance(arg, Column):
                local = validate_column(arg, context)
                valid = valid and local
        return valid

    def validate_literal(lit: Literal, context: QueryContext) -> bool:
        if isinstance(lit.value, str) and lit.value in environment.concepts:
            return True
        elif isinstance(lit.value, Calculation):
            return validate_calculation(lit.value, context)
        return True

    def validate_column(col: Column, context: QueryContext) -> bool:
        valid = False
        if (
            col.name not in select
            and col.name not in environment.concepts
            and not col.calculation
        ):
            recommendations = None
            try:
                environment.concepts[col.name]
            except UndefinedConceptException as e:
                recommendations = e.suggestions
            if context == QueryContext.FILTER:
                hint = "add another nested column with a calculation to define it in this filter"
            else:
                hint = "add a select to calculate it"
            if recommendations:
                errors.append(
                    f"{col.name} in {context} is not a valid field or previously defined by you; if this is a new metric you need, {hint}. Is one of these the correct spelling? {recommendations}?",
                )
            else:

                errors.append(
                    f"{col.name} in {context} is not a valid field or previously defined by you. If this is a new metric you need,  {hint}. If you misspelled a field, potential matches are {difflib.get_close_matches(col.name, environment.concepts.keys(), 3, 0.4)}",
                )
        elif col.name in environment.concepts:
            valid = True
            if col.calculation:
                errors.append(
                    f"{col.name} in {context} is a predefined field and should not have a calculation; check that you are using the field as is, without any additional transformations.",
                )
        elif col.calculation:

            if not is_valid_function(col.calculation.operator):
                errors.append(
                    f"{col.name} Column definition in {context} does not use a valid function (is using {col.calculation.operator}); check that you are using ONLY a valid option from this list: {sorted([x for x in FunctionType.__members__.keys() if x not in COMPLICATED_FUNCTIONS]) }. If the column requires no transformation, drop the calculation field.",
                )
            else:
                valid = True
            if col.calculation.operator in ["AVG", "SUM", "MIN", "MAX", "COUNT"]:
                for arg in col.calculation.arguments:
                    if isinstance(arg, Column):
                        if arg.name in environment.concepts:
                            dtype = environment.concepts[arg.name].datatype
                            if dtype == DataType.STRING:
                                errors.append(
                                    f"Aggregate function {col.calculation.operator} in {context} is being used on a string field {arg.name}; do you need a calculation, or are you using the wrong field?",
                                )
            if col.calculation.over:
                for x in col.calculation.over:
                    local = validate_column(x, context)
                    valid = valid and local

            for arg in col.calculation.arguments:
                if isinstance(arg, Column):
                    local = validate_column(arg, context)
                    valid = valid and local
        if valid:
            if context == QueryContext.SELECT:
                select.add(col.name)
            elif context == QueryContext.FILTER:
                filtered_on.add(col.name)
        return valid

    for x in parsed.output_columns:
        validate_column(x, QueryContext.SELECT)

    if parsed.order:
        for y in parsed.order:
            if y.column_name not in select:
                errors.append(
                    f"{y.column_name} being ordered by is not in output_columns, add if needed.",
                )
    if parsed.filtering:
        root = parsed.filtering.root
        for val in root.values:
            if isinstance(val, Column):
                validate_column(val, QueryContext.FILTER)

            elif isinstance(val, NLPConditions):
                for subval in [val.left, val.right]:
                    if isinstance(subval, Column):
                        validate_column(subval, QueryContext.FILTER)
                    elif isinstance(subval, Literal):
                        validate_literal(subval, QueryContext.FILTER)
                if isinstance(val.left, Column) and isinstance(val.right, Column):
                    if val.left.name == val.right.name:
                        errors.append(
                            f"Comparison {val} has the same column name on both sides; this is a meaningless comparison. Check that you are comparing two different fields.",
                        )

    if errors:
        return {
            "status": "invalid",
            "errors": {f"Error {idx}: {error}" for idx, error in enumerate(errors)},
        }, parsed
    tips = [
        f'No validation errors - looking good! Just double check you are outputting only (and all of) the required info from the original prompt, and submit it! (Remember that a column used just for filtering should only exist in filtering; do ) Prompt: "{prompt}"!'
    ]
    for all_address in select.union(filtered_on):
        if all_address in environment.concepts:
            concept = environment.concepts[all_address]
            if concept.metadata.description:
                tips.append(
                    f'For {all_address}, reminder that the field description is "{concept.metadata.description}". Double check any filtering on this field matches the described format! (this is a reminder, not an error)'
                )
    VALIDATION_CACHE[prompt] = parsed
    return {"status": VALID_STATUS, "tips": tips}, parsed


def validation_error_to_string(err: ValidationError):
    # inject in new context on failed answer
    raw_error = str(err)
    errors = err.errors()
    if "filtering.root." in raw_error:
        missing = []
        path_freq: dict[str, int] = defaultdict(lambda: 0)
        path_map: dict[str, list[str]] = defaultdict(lambda: [])
        for validation_error in errors:
            e: ErrorDetails = validation_error
            if e["type"] == "missing":
                path = "".join([str(x) for x in e["loc"][:-1]])
                path_freq[path] += 1
                message = f'Missing "{e["loc"][-1]}" in this JSON object you provided: {json.dumps(e["input"], indent=4)}. You may also be using the wrong object. Double check Literal and Column formats.'
                path_map[path].append(message)
                missing.append(message)
            elif e["type"] == "extra_forbidden":
                path = "".join([str(x) for x in e["loc"][:-1]])
                path_freq[path] += 1
                message = f'Extra invalid key "{e["loc"][-1]}" in this JSON object you provided: {json.dumps(e["input"], indent=4)}'
                path_map[path].append(message)
                missing.append(message)
            else:
                missing_path = ".".join(
                    [
                        str(v)
                        for v in e["loc"]
                        if str(v)
                        not in [
                            "NLPComparisonGroup",
                            "NLPConditions",
                            "Literal",
                            "Column",
                        ]
                    ]
                )
                missing.append(missing_path)
        min_path = min(path_freq, key=path_freq.get)  # type: ignore
        locations = path_map[min_path]
        raw_error = f"Syntax error in your filtering clause. Check that you only use valid keys for Literal, Comparison, and Column objects, and have all required keys. Look at these locations: {locations}"
    else:
        for e in errors:
            if e["type"] == "missing":
                raw_error = f'Missing {e["loc"][-1]} in {e["input"]}. You might have incorrect JSON formats or the key is actually missing. Double check Literal and Column formats.'
    return raw_error


def validate_response(
    environment: Environment,
    prompt: str,
    output_columns: List[Column],
    filtering: Optional[FilterResultV2] = None,
    order: Optional[list[OrderResultV2]] = None,
    limit: int | None = None,
) -> tuple[dict, InitialParseResponseV2 | None]:

    base: dict[str, list[Column] | FilterResultV2 | list[OrderResultV2] | int] = {
        "output_columns": output_columns
    }
    if not output_columns:
        raise ToolException(
            "A request to validate should include at least one column. The input should be the exactly same as the input for your final answer. Generate your answer, call this tool again, and then submit it if it passes."
        )
    if filtering:
        base["filtering"] = filtering
    if order:
        base["order"] = order
    if limit:
        base["limit"] = limit
    return validate_query(
        base,
        environment=environment,
        prompt=prompt,
    )


class ValidateResponseInterface(BaseModel):
    # deliberately permissive interface
    # so that we can handle the error inside the tool
    output_columns: List[dict]
    filtering: Optional[dict] = None
    order: Optional[list[dict]] = None
    limit: int | None = None
